# Copyright (C) 2003-2005 Kouichirou Eto, All rights reserved.

$LOAD_PATH << "../../lib" unless $LOAD_PATH.include?("../../lib")
require "qwik/wikidb"
require "qwik/fsdb"
require "qwik/util-string"

module Qwik
  class PageExistError < StandardError; end
  class PageCollisionError < StandardError; end

  class Pages
    include Enumerable

    FIRST_PAGE_ID = "1"

    def initialize(config, dir)
      @config = nil
      dir = dir.path if dir.is_a? String
      spath = config.super_dir.path
      if config.db == "fsdb"
	@db = FileSystemDB.new(dir, spath)
      elsif config.db == "bdb"
	@db = BerkeleyDB.new(dir, spath)
      else
	raise "Unknown data base type."
      end
      @pages = {}
    end
    attr_reader :db

    def backupdb
      @db.backupdb
    end

    def close
      @db.close
    end

    def create(k) # create new page
      raise PageExistError if baseexist?(k)
      @pages[k] = Page.new(@config, self, k)
    end

    def create_new
      create(get_new_id)
    end

    def baseexist?(k)
      @db.baseexist?(k)
    end

    def exist?(k)
      return true if @pages[k]
      return @db.exist?(k)
    end

    def [](k)
      return @pages[k] if @pages[k]
      return @pages[k] = Page.new(@config, self, k) if @db.exist?(k)
      nil # no such page
    end

    def delete(k)
      self[k].delete
      @pages[k] = nil
    end

    def to_a(all=nil)
      ar = []
      each(all){|page|
	ar << page
      }
      ar
    end

    def each(all=nil)
      @db.each {|key|
	next if !all && /\A_/ =~ key
	page = self[key]
	next if page.nil? # what?
	yield(page)
      }
    end

    def each_all(&b)
      each(true, &b)
    end

    def keys
      map {|page| page.key }
    end

    def title_list
      sort_by {|page|
	page.get_title.downcase
      }
    end

    def date_list
      sort_by {|page|
	page.mtime
      }
    end

    def find_title(title)
      find {|page|
	page.get_title == title
      }
    end

    def get_new_id
      numkeys = keys.map {|k| /\A[0-9]+\z/ =~ k ? k.to_i : 0 }
      return (numkeys.max + 1).to_s if numkeys.max
      return FIRST_PAGE_ID
    end

    def last_page_time
      @db.last_page_time
    end

    # QuickML support
    def last_article_time
      @db.last_article_time
    end

    # page title support
    def pagetitle
      @pagetitle = PageTitle.new(@config, self) unless defined? @pagetitle
      @pagetitle
    end

    def get_by_title(title)
      page = self[title] rescue nil
      return page if page

      key = pagetitle.hash[title]
      return self[key[0]] if key

      page = self.find_title(title)
      return page if page

      nil
    end
  end

  class Page
    def initialize(config, pages, key)
      @pages = pages
      @key = key.to_s
      @db = @pages.db
      @db.create(@key)
    end
    attr_reader :key

    def inspect
      "#<Page:"+@key+">"
    end

    def url
      @key+".html"
    end

    def mtime
      @db.mtime(@key)
    end

    def get
      @db.get(@key)
    end
    alias load get

    def put(v, md5=nil)
      raise PageCollisionError if md5 && load.md5hex != md5
      @db.put(@key, v)
    end
    alias store put

    def add(v)
      str = get
      addstr = ""
      addstr += "\n" if !(str.empty? || str[-1] == ?\n)
      addstr += v
      addstr += "\n" if !(v.empty? || v[-1] == ?\n)
      @db.add(@key, addstr)
    end
    alias append add

    def delete
      @db.delete(@key)
    end

    def touch
      @db.touch(@key)
    end

    def <=>(o)
      self.key <=> o.key
    end

    def get_body
      ar = []
      first_line = true
      load.each_line {|line|
	next if /\A#/ =~ line
	if first_line
	  first_line = false
	  if line[0] == ?* && line[1] != ?*
	    # this is title
	  else
	    ar << line
	  end
	else
	  ar << line
	end
      }
      ar.join
    end

    def get_title
      t = parse_first_line
      return t if t

      rh = @pages.pagetitle.rev_hash
      t = rh[@key]
      return t if t

      return @key
    end

    def parse_first_line
      line = get_first_line
      return nil unless /\A\*(.+)\z/s =~ line # line does not begin with * ?

      title = $1
      return nil if /\A\*/s =~ title # only h2 level

      title = title.strip
      return nil if title.length == 0
      return title
    end

    def old_get_first_line
      load.each_line {|line|
	next if /\A#/ =~ line
	return line.chomp
      }
      ""
    end

    def new_get_first_line
      str = load
      while true
	i = str.index(?\n)
	return str if i.nil?
	line = str[0..i]
	line.chop!
	rest = str[(i+1)..-1]
	if line[0] == ?#
	    str = rest
	  next
	end
	return line
      end
    end
    alias get_first_line old_get_first_line
#   alias get_first_line new_get_first_line

    def wikidb
      @wikidb = WikiDB.new(self) unless defined? @wikidb
      @wikidb
    end
  end

  class PageTitle
    KEY = "_PageTitle"

    def initialize(config, site)
      @site = site
    end

    def add(t, k)
      db.add(t, k)
    end

    def hash
      db.hash
    end

    def rev_hash
      db.rev_hash
    end

    private
    def db
      page = @site[KEY]
      page = @site.create(KEY) if page.nil?
      page.wikidb
    end
  end
end

if $0 == __FILE__
  require "test/unit"
  $LOAD_PATH << "../../lib" unless $LOAD_PATH.include?("../../lib")
  require "qwik/config"
  require "qwik/util-path"
  require "qwik/bdb"
  require "qwik/qp"
  require "qwik/memory"
  require "t/module-session"
  $debug = true
end

if $debug
  class TC_Pages < Test::Unit::TestCase
    include TestSession

    def assert_first_line(page, e, s)
      page.store(s)
      assert_equal(e, page.get_first_line)
    end

    def assert_title(page, e, s)
      page.store(s)
      assert_equal(e, page.get_title)
    end

    def assert_no_title(page, s)
      page.store(s)
      assert_equal(page.key, page.get_title)
    end

    def test_page
      @pages = @site.get_pages

      assert_equal(false, @pages.exist?("1"))
      assert_equal(nil, @pages["1"])

      page = @pages.create("1")
      assert_instance_of(Qwik::Page, page)
      assert_equal(true, @pages.exist?("1"))
      assert_equal("", page.load)
      page.store("t")
      assert_raises(Qwik::PageExistError){ @pages.create("1") }

      page.store("test1")
      assert_equal("test1", page.load)

      page = @pages["1"]
      assert_instance_of(Qwik::Page, page)
      assert_equal("test1", page.load)
      page.store("test2")
      assert_equal("test2", page.load)

      # test_first_line
      assert_first_line(page, "line1", "line1")
      assert_first_line(page, "l1",    "l1\nl2\n")
      assert_first_line(page, "",      "\nline2\n")
      assert_first_line(page, "l1",    "# c\nl1\nl2\n")

      # test_url
      assert_equal("1.html", page.url)

      # test_title
      assert_title(page, "t", "*t")
      assert_no_title(page, "")
      assert_title(page, "t", "*t ")
      assert_title(page, "t", "* t")
      assert_title(page, "t", "* t ")
      assert_title(page, "a b", "*a b")
      assert_title(page, "t", "* t")
      assert_title(page, "a b", "* a b")
      assert_title(page, "- t", "*- t")
      assert_title(page, "", "*")
      assert_title(page, "t", "*t")
      assert_no_title(page, "**t")
      assert_title(page, "*t", "* *t")
      assert_no_title(page, "-t")
      assert_title(page, "", "*")
      assert_title(page, "R", "* R")

      #page.store("*t");	assert_equal("t", page.get_title)
      #page.add("**t");	assert_equal("1", page.get_title)

      assert_title(page, "R", "* R\n{{mail(user@e.com)\n\n\n}}\n")

      # test_key
      assert_equal("1", page.key)

      @pages.each {|page|
	assert_instance_of(Qwik::Page, page)
	assert_instance_of(String, page.key)
      }

      page3 = @pages.create("_t") # test begin with _
      page3.store("t")
      found = false
      @pages.each {|page|
	found = true if page.key[0] == ?_
      }
      assert_equal(false, found)
      assert_equal(false, @pages.any? {|page| page.key[0] == ?_ })

      @pages.each(true) {|page|
	found = true if page.key[0] == ?_
      }
      assert_equal(true, found)

      found = false
      @pages.each_all {|page|
	#qp page.key
	found = true if page.key[0] == ?_
      }
      assert_equal(true, found)

      # test_mtime
      assert_instance_of(Time, page.mtime)

      # test_last_page_time
      assert_instance_of(Time, @pages.last_page_time)

      # test_add
      page = @pages["1"]

      page.store("t1")
      page.add("t2")
      assert_equal("t1\nt2\n", page.load)

      page.store("")
      page.add("t2")
      assert_equal("t2\n", page.load)

      page.store("t1")
      page.add("")
      assert_equal("t1\n", page.load)

      page.delete
      page.add("")
      assert_equal("", page.load)

      # test_substitute
      page = @pages["1"]

      page.store("t")
      str = page.load
      assert_equal("t", str)
      str = str.sub(/t/, "s")
      assert_equal("s", str)
      assert_equal("t", page.load)

      # test_delete
      @pages.delete("1")
      assert_equal(false, @pages.exist?("1"))

      # test_store_md5
      page.store("t1")
      assert_raises(Qwik::PageCollisionError){ page.store("t1", "1") }
      md5 = page.load.md5hex
      page.store("t2", md5)
      assert_equal("t2", page.load)
      page.add("t3")
      assert_equal("t2\nt3\n", page.load)
#      assert_equal("t2t3", page.load)

      # test_super_pages
      assert_equal(true, @pages.exist?("_SideMenu"))
      page = @pages["_SideMenu"]
      assert_equal("menu", page.get_title)
      assert_equal("TCgj[", @pages["_SiteMenu"].get_title)
      assert_instance_of(Time, @pages["_SiteMenu"].mtime)

      assert_equal(true, @pages.exist?("_SideMenu")) # exist
      assert_equal(false, @pages.baseexist?("_SideMenu")) # non exist
      page = @pages.create("_SideMenu") # can create
      assert_equal(false, @pages.baseexist?("_SideMenu")) # not yet
      page.store("t")
      assert_equal(true, @pages.baseexist?("_SideMenu")) # become true here

      # test_delete
      assert_instance_of(Time, @pages.last_page_time)

      page = @pages["1"]
      page.store("* a")
      assert_instance_of(Time, page.mtime)
      assert_instance_of(Time, @pages.last_page_time)

      page.delete
      assert_instance_of(Time, page.mtime)
      assert_instance_of(Time, @pages.last_page_time)
      assert_equal(0, page.mtime.to_i)

      # test_get_new_id
      page = @pages.create("0");	page.store("t")
      assert_equal("1", @pages.get_new_id)
      page = @pages.create(@pages.get_new_id);	page.store("t")
      assert_equal(true, @pages.exist?("1"))
      assert_equal("2", @pages.get_new_id)
      page = @pages.create("t");	page.store("t")
      assert_equal("2", @pages.get_new_id)
      assert_equal("2", @pages.get_new_id)
      page = @pages.create("4");	page.store("t")
      assert_equal("5", @pages.get_new_id)

      # test_get_body
      @pages = @site.get_pages
      page = @pages["1"]
      assert_title_body(page, "t", "b", "* t\nb")
      assert_title_body(page, "t", "b", "# c\n* t\nb")
      assert_title_body(page, "1", "b1\nb2", "b1\nb2")
      assert_title_body(page, "1", "** t\nb", "** t\nb")

      # test_page_with_null_title
      @pages = @site.get_pages
      page = @pages["1"]
      page.store("*t");	assert_equal("t", page.get_title)
      page.store("*");	assert_equal("1", page.get_title)
      page.store("* ");	assert_equal("1", page.get_title)
      page.store("*  ");	assert_equal("1", page.get_title)
    end

    def test_compare
      @pages = @site.get_pages
      @pages.erase_all
      a = @pages.create("a")
      b = @pages.create("b")
      pages = @pages.sort
      assert_equal("a", pages[0].get_title)
      assert_equal("b", pages[1].get_title)
    end

    def test_list
      @pages = @site.get_pages

      @pages.create("t1").store("")
      @pages.create("t2").store("")
      @pages.create("1").store("t3")
      @pages.create("2").store("t4")

      @pages.each {|page|
	assert_instance_of(Qwik::Page, page)
      }

      tlist = @pages.title_list
      assert_equal("1", tlist[0].key)
      assert_equal("2", tlist[1].key)
      assert_equal("t1", tlist[2].key)
      assert_equal("t2", tlist[3].key)
      tlist.each {|page|
	assert_instance_of(String, page.get_title)
      }
    end

    def test_last_article_time
      @pages = @site.get_pages
      a = @pages.last_article_time
      assert_equal(nil, a)
      page = @pages.create_new
      a = @pages.last_article_time
      assert_instance_of(Time, a)
    end

    def test_enumerable
      @pages = @site.get_pages

      #test_find_title
      @pages.create_new.store("* ")
      @pages.create_new.store("* ")
      @pages.create_new.store("* ")

      #page = @pages.find_title("")
      page = @pages.get_by_title("")
      assert_equal("2", page.key)
      assert_equal(["1", "2", "3"], @pages.keys)
    end

    def test_recent_list
      @pages = @site

      # Since this test takes long time, ignore this if under test suite.
      return if $0 != __FILE__
      @pages.create("t1").store("t");	sleep 1
#      @pages.create("t2").store("t");	sleep 1
      @pages.create("1").store("t3")
#;	sleep 1
#      @pages.create("2").store("t4")
      dlist = @pages.date_list
      assert_equal("t1", dlist[0].key)
#      assert_equal("t2", dlist[1].key)
#      assert_equal("1",  dlist[2].key)
      assert_equal("1",  dlist[1].key)
#      assert_equal("2",  dlist[3].key)
      dlist.each {|page|
	assert_instance_of(Time, page.mtime)
      }
    end

    def nutest_cache
      @pages = @site

      # Since this test takes long time, ignore this if under test suite.
      return if $0 != __FILE__

      length, repeat = 100, 100

      page = @pages.create_new
      page.store("a" * length)
      sleep 1
      repeat.times {
	str = page.load
      }
    end

    def test_backup
      @pages = @site

      page = @pages.create("1")
      page.store("t")
      assert_equal("t", str = page.load)

      bdb = @pages.backupdb
#      assert_instance_of(Qwik::BackupDB, bdb)
      bdb.each_by_key("1") {|v, time|
#	assert_instance_of(Pathname, f)
	assert_instance_of(String, v)
	assert_instance_of(Time, time)
#	str = f.read
#	assert_equal("t", str)
	assert_equal("t", v)
      }

      return if $0 != __FILE__ # just for UNIT TEST

      sleep 1
      page.store("t2")

      bdb.each_by_key("1") {|v, time|
	assert_match(/\At/, v)
      }
    end

    def assert_title_body(page, title, body, s)
      @pages = @site

      page.store(s)
      assert_equal(title, page.get_title)
      assert_equal(body,  page.get_body)
    end

    def test_touch
      @pages = @site

      page = @pages["FrontPage"]
      org = page.load
      assert_match(/FrontPage/, page.load)

      page.touch
      assert_match(/FrontPage/, page.load)
      assert_equal(org, page.load)
    end
  end
end
