# Copyright (C) 2003-2005 Kouichirou Eto, All rights reserved.

$LOAD_PATH << "../../lib" unless $LOAD_PATH.include?("../../lib")
require "qwik/util-string"
require "qwik/tokenizer"
require "qwik/htmlparser"
#require "qwik/emode-parser"
require "qwik/wabisabi-valid"

module Qwik
  class TextParser # make a tree
    def parse(str)
      str = str.map {|line| line.chompp }.join("\n")

      if emode?(str) # special mode
	sar = preprocess_emode(str)
	str = sar.join("\n")
      end

      tokens = TextTokenizer.new.tokenize(str)
      tree = make_tree(tokens)
      return tree
    end

    def emode?(str)
      /\A===========================================================+/ =~ str
    end

    def preprocess_emode(str)
      return EmodePreProcessor.new.preprocess(str)
    end

    def make_tree(tokens)
      ar = []

      tokens.each {|token|

	tag = token.first
	case tag

	when :h2, :h3, :h4, :h5, :h6
	  ar << [tag, *inline(token[1])]

	when :ul, :ol
	  lu = ar
	  token[1].times {
	    if lu.last && lu.last.is_a?(Array) && lu.last.first == tag
	      lu = lu.last
	    else
	      a = [tag]
	      lu << a
	      lu = a
	    end
	  }
	  lu << [:li, *inline(token[2])]

	when :pre, :blockquote
	  if ar.last && ar.last.first == tag
	    ar.last[1] << "\n"+token[1]
	  else
	    ar << [tag, token[1]]
	  end

	when :dl
	  a = []
	  a << [:dt, *inline(token[1])] if token[1]
	  a << [:dd, *inline(token[2])] if token[2]
	  if ar.last && ar.last.first == tag
	    ar.last << a[0]
	    ar.last << a[1] if a[1]
	  else
	    ar << [tag, *a]
	  end

	when :table
	  token.shift
	  table_ar = token.map {|td|
	    inline_ar = inline(td)
	    inline_ar << "" if inline_ar.empty?
	    [:td] + inline_ar
	  }
	  tr = [:tr] + table_ar
	  if ar.last && ar.last.first == tag
	    ar.last << tr
	  else
	    ar << [tag, tr]
	  end

	when :text
	  line_ar = inline(token[1])
	  if ar.last && ar.last.is_a?(Array) && ar.last.first == :p
	    ar[-1] << "\n" if ar.last.last != [:br]
	    ar[-1] += line_ar
	  else
	    ar << [:p, *line_ar]
	  end

	when :empty
	  ar << ["\n"]

	when :plugin
	  ar << parse_plugin(token)

	when :html
	  ar << parse_html(token)

	else
	  ar << token
	end
      }

      nar = []
      ar.each {|block|
	tag = block.first
	case tag
	when :blockquote
	  str = block[1]
	  nar << [tag, *parse(str)]
	when :p
	  while block[1] == [:br]
	    block.delete_at(1)
	  end
	  while block.last == [:br]
	    block.pop
	  end
	  nar << block
	else
	  nar << block
	end
      }

      return nar
    end

    def inline(str)
      return InlineParser.new.parse(str)
    end

    def parse_plugin(token)
      method = token[1].to_s
      param = token[2].to_s
      data = token[3].to_s.chomp
      return [:plugin, {:method=>method, :param=>param}, data]
    end

    def parse_html(token)
      str = token[1]
      wabisabi = HtmlParser.new.parse(str)

      v = wabisabi.check_valid
      if v == true
	return [:html, *wabisabi]
      else
	return [:div, {:class=>"error"}, "can not use ["+v.to_s+"]"]
      end
    end
  end

  class InlineParser
    include InlinePatternModule

    def parse(str)
      if str.is_a? Array
	return [str]
      end
      tokenizer = InlineTokenizer.new
      tokens = tokenizer.tokenize(str)
      tree = make_tree(tokens)
      return tree
    end

    def make_tree(line_ar)
      table = {
	:"'''"=>:strong,
	:"''"=>:em,
	:"=="=>:del,
      }

      root = []
      stack = []
      stack << root
      in_tag = {}

      line_ar.each {|t|
	if t.is_a? Symbol
	  if in_tag[t]
	    in_tag[t] = nil
	    if stack.last.length == 1 # bad hack
	      stack.last << ""
	    end
	    stack.pop if 1 < stack.length
	  else
	    in_tag[t] = true
	    ar = [table[t]]
	    stack.last << ar
	    stack << ar
	  end

	else
	  if t.is_a? Array
	    elem = t.first
	    case elem
	    when :ref
	      stack.last << parse_ref(t[1])
	    when :plugin
	      stack.last << parse_plugin(t)
	    when :url
	      stack.last << [:a, {:href=>t[1]}, t[1]]
	    else
	      # do nothing
	    end

	  else
	    stack.last << t
	  end
	end
      }
      return root
    end

    def parse_plugin(t)
      method = t[1].to_s
      param = t[2].to_s
      data = t[3].to_s
      case method
      when "br"
	return [:br] # shortcut
      end
      return [:plugin, {:method=>method, :param=>param}, data]
    end

    def parse_ref(ref)
      unless /\A(?:(.+)\|)?(.+)\z/ =~ ref
	raise "unknown ref"
      end

      s = $1
      href = $2

      if /\A#{URL}/ =~ href
	if /\.(?:jpg|jpeg|png|gif)\z/i =~ href && s.nil?
	  s = href
	  return [:img, {:src=>href, :alt=>s}] # embed the image
	end
	s = href if s.nil?
	return [:a, {:href=>href}, s]

      elsif /\A(.+?):(.+)\z/ =~ href
	disp = s ? s : "#{$1}:#{$2}"
	return [:plugin, {:method=>"interwiki", :param=>"#{$1}:#{$2}"}, disp]

      end

      s = href.dup if s.nil?
      h = href.to_s # do not escape it for now
      h = h+".html" unless h.include?(".")
      return [:a, {:href=>h}, s]
    end

  end
end

if $0 == __FILE__
  require "test/unit"
  $LOAD_PATH << "../../lib" unless $LOAD_PATH.include?("../../lib")
  require "qwik/qp"
  #require "qwik/wabisabi.rb"
  require "qwik/format-xml"
  $debug = true
end

if $debug
  class TC_TextParser < Test::Unit::TestCase
    def assert(e, str)
      tokens = Qwik::TextParser.new.parse(str)
      assert_equal(e, tokens)
    end

    def test_all
      assert([], "")
      assert([["\n"]], "#t")
      assert([[:h2, "t"]], "*t")
      assert([[:h2, "t"], [:p, "t"]], "*t\nt")
      assert([[:ul, [:li, "t"]]], "-t")
      assert([[:ul, [:ul, [:li, "t"]]]], "--t")
      assert([[:ul, [:ul, [:ul, [:li, "t"]]]]], "---t")
      assert([[:ul, [:li, "t"]], [:p, "t"]], "-t\nt")
      assert([[:ul, [:li, "t"], [:li, "t"]]], "-t\n-t")
      assert([[:ul, [:li, "t"], [:ul, [:li, "t"]]]], "-t\n--t")
      assert([[:ul, [:ul, [:li, "t"]], [:li, "t"]]], "--t\n-t")
      assert([[:ul, [:ul, [:li, "t"]], [:li, "t"], [:ul, [:li, "t"]]]],
	     "--t\n-t\n--t")
      assert([[:ol, [:li, "t"]]], "+t")
      assert([[:ul, [:li, "t"]], [:ol, [:li, "t"]]], "-t\n+t")
      assert([[:blockquote, [:p, "t"]]], ">t")
      assert([[:blockquote, [:p, "s", "\n", "t"]]], ">s\n>t")
      assert([[:blockquote, [:ul, [:li, "s"], [:li, "t"]]]], ">-s\n>-t")
      assert([[:dl, [:dt, "dt"],[:dd, "dd"]]], ":dt:dd")
      assert([[:dl, [:dt, "dt"]]], ":dt")
      assert([[:dl]], ":")
      assert([[:dl]], "::")
      assert([[:dl, [:dd, "dd"]]], "::dd")
      assert([[:dl, [:dt, "dt"], [:dd, "dd"], [:dt, "dt2"], [:dd, "dd2"]]],
	     ":dt:dd\n:dt2:dd2")
      assert([[:dl, [:dt, "dt"], [:dd, "dd"], [:dd, "dd2"]]], ":dt:dd\n::dd2")
      assert([[:pre, "t"]], " t")
      assert([[:pre, "s\nt"]], " s\n t")

      # test_table
      assert([[:table, [:tr, [:td, "t"]]]], ",t")
      assert([[:table, [:tr, [:td, "t1"], [:td, "t2"]]]], ",t1,t2")
      assert([[:table, [:tr, [:td, ""], [:td, "t2"]]]], ",,t2")
      assert([[:table, [:tr, [:td, "s"]], [:tr, [:td, "t"]]]], ",s\n,t")
      assert([[:table, [:tr, [:td, "s1"], [:td, "s2"]],
		 [:tr, [:td, "t1"], [:td, "t2"]]]], ",s1,s2\n,t1,t2")

      assert([[:plugin, {:method=>"t", :param=>""}, ""]], "{{t}}")
      assert([[:plugin, {:method=>"t", :param=>"a"}, ""]], "{{t(a)}}")
      assert([[:plugin, {:method=>"t", :param=>""}, "s"]], "{{t\ns\n}}")
      assert([[:plugin, {:method=>"t", :param=>""}, "s1\ns2"]],
	     "{{t\ns1\ns2\n}}")
      assert([[:pre, "s\n"]], "{{{\ns\n}}}")
      assert([[:pre, "s\nt\n"]], "{{{\ns\nt\n}}}")
      assert([[:p, "t"]], "t")
      assert([[:p, "s", "\n", "t"]], "s\nt")
      assert([[:p, "s"], ["\n"], [:p, "t"]], "s\n\nt")

      assert([[:dl, [:dt, "dt1"], [:dd, "dd1"]], [:p, "p1"],
	       [:dl, [:dd, "dd2"]]], ":dt1:dd1\np1\n::dd2")

      # test_multiline
      assert([[:p, "s", "\n", "t"]], "s\nt")
      assert([[:p, "s", [:br], "t"]], "s~\nt~")
      assert([[:p, "s"], ["\n"], [:p, "t"]], "s\n\nt")
      assert([[:p, "s", [:br], [:br], "t"]], "s~\n~\nt~")

      str = <<"EOT"
p1~
~
> b1~
> b2~
> ~
> > bb1~
> > bb2~
> ~
> b3~
> b4~
~
p2~
EOT
      assert([[:p, "p1"],
	       [:blockquote,  [:p, "b1", [:br], "b2"],
		 [:blockquote, [:p, "bb1", [:br], "bb2"]],
		 [:p, "b3", [:br], "b4"]],
	       [:p, "p2"]],
	     str)

      # test_ul
      assert([[:ul, [:li, "t"]]], "-t")
      assert([[:ul, [:li, "t1"], [:li, "t2"]]], "-t1\n-t2")
      assert([[:ul, [:li, "t1"]], ["\n"], [:ul, [:li, "t2"]]], "-t1\n\n-t2")
      assert([["\n"], [:ul, [:li, "t1"], [:li, "t2"]]], "\n-t1\n-t2")

      # test_multiline_pre
      assert([[:pre, "t"]], " t")
      assert([[:pre, "t1\nt2"]], " t1\n t2")
      assert([[:pre, "t1"], ["\n"], [:pre, "t2"]], " t1\n\n t2")

      # test_html
      assert([[:html, "t\n"]], "<html>\nt\n</html>")
      assert([[:div, {:class=>"error"}, "can not use [script]"]],
	     "<html>\n<script>t</script>\n</html>")
    end
  end

  class TC_InlineParser < Test::Unit::TestCase
    def assert(e, str)
      tokens = Qwik::InlineParser.new.parse(str)
      assert_equal(e, tokens)
    end

    def test_all
      assert([], "")
      assert(["t"], "t")
      assert([[:em, "t"]], "''t''")
      assert([[:strong, "t"]], "'''t'''")
      assert(["'"], "'")
      assert(["'", "t"], "'t")
      assert([[:del, "t"]], "==t==")
      assert(["="], "=")
      assert(["=", "t"], "=t")
      assert([[:a, {:href=>"t.html"}, "t"]], "[[t]]")
      assert([[:a, {:href=>"t.html"}, "s"]], "[[s|t]]")
      assert([[:a, {:href=>"http://e.com/"}, "e"]], "[[e|http://e.com/]]")
      assert([[:plugin, {:method=>"interwiki", :param=>"w:p"}, "w:p"]],
	     "[[w:p]]")
      assert(["]"], "]")
      assert(["[", "t", "]"], "[t]")
      assert([[:plugin, {:method=>"t", :param=>""}, ""]], "{{t}}")
      assert([[:plugin, {:method=>"t", :param=>"a"}, ""]], "{{t(a)}}")
      assert(["{", "t", "}"], "{t}")
      assert([[:a, {:href=>"http://e.com/"}, "http://e.com/"]], "http://e.com/")
      assert([[:a, {:href=>"https://e.com/"}, "https://e.com/"]],
	     "https://e.com/")
      assert(["t ", [:a, {:href=>"http://e.com/"}, "http://e.com/"], " a"],
	     "t http://e.com/ a")
      assert(["a ", [:em, "b"], " ", [:strong, "c"], " ", [:del, "d"], " ",
	       [:a, {:href=>"e.html"}, "e"], " ",
	       [:plugin, {:method=>"f", :param=>""}, ""], " ",
	       [:a, {:href=>"http://e.com/"}, "http://e.com/"], " g"],
	     "a ''b'' '''c''' ==d== [[e]] {{f}} http://e.com/ g")
      assert(["a ",[:em,"b ",[:strong,"c"]," d"]," e"], "a ''b '''c''' d'' e")
      assert([[:em, "a", [:strong, "b", [:del, "c"]]]], "''a'''b==c") # uum...
      assert([[:strong, ""]], "''''''")
      assert(["a", [:em, "b", [:strong, "c"], "d"], "e"], "a''b'''c''d'''e")
      assert(["a", [:del, "b", [:em, "c"], "d"], "e"], "a==b''c==d''e")

      # test_img
      assert([[:img, {:alt=>"http://e.com/t.jpg", :src=>"http://e.com/t.jpg"}]],
	     "[[http://e.com/t.jpg]]")
      assert([[:a, {:href=>"http://e.com/t.jpg"}, "m"]],
	     "[[m|http://e.com/t.jpg]]")

      # test_security
      assert([[:a, {:href=>"t.html"}, "t"]], "[[t]]")
      assert([[:a, {:href=>"http://e.com/"}, "http://e.com/"]],
	     "[[http://e.com/]]")
      assert([[:plugin, {:param=>"javascript:t", :method=>"interwiki"},
		 "javascript:t"]],
	     "[[javascript:t]]") # Treated as InterWiki
      assert([[:a, {:href=>"&{alert('hello');};.html"}, "&{alert('hello');};"]],
	     "[[&{alert('hello');};]]")
    end
  end

  class TC_InlineParser_parse_ref < Test::Unit::TestCase
    def assert(e, str)
      assert_equal(e, Qwik::InlineParser.new.parse_ref(str))
    end

    def test_all
      assert([:a, {:href=>"t.html"}, "t"], "t")
      assert([:a, {:href=>"t.html"}, "s"], "s|t")
      assert([:a, {:href=>"http://e.com/"}, "http://e.com/"], "http://e.com/")
      assert([:a, {:href=>"http://e.com/"}, "s"], "s|http://e.com/")
      assert([:plugin, {:param=>"s:t", :method=>"interwiki"}, "s:t"], "s:t")

      # test_plugin_ref
      assert([:a, {:href=>".attach"}, ".attach"], ".attach")
      assert([:a, {:href=>".attach/t.txt"}, ".attach/t.txt"], ".attach/t.txt")
      assert([:a, {:href=>".attach/s t.txt"}, ".attach/s t.txt"],
	     ".attach/s t.txt")

      # test_img
      assert([:img, {:alt=>"http://e.com/t.jpg", :src=>"http://e.com/t.jpg"}],
	     "http://e.com/t.jpg")
      assert([:a, {:href=>"http://e.com/t.jpg"}, "s"],
	     "s|http://e.com/t.jpg")

      # test security
      assert([:a, {:href=>"&{alert('hello');};.html"}, "&{alert('hello');};"],
	     "&{alert('hello');};")
    end
  end

  class TC_TextParserWithFormat < Test::Unit::TestCase
    def assert(e, str)
      tokens = Qwik::TextParser.new.parse(str)
      assert_equal(e, tokens.rb_format_xml(-1, -1))
    end

    def test_all
      assert("", "")
      assert("<h2>t</h2>", "*t")
      assert("<h2>t</h2><p>t</p>", "*t\nt")
      assert("<ul><li>t</li></ul>", "-t")
      assert("<blockquote><p>s\nt</p></blockquote>", ">s\n>t")
      assert("<blockquote><ul><li>t1</li><li>t2</li></ul></blockquote>",
	     ">-t1\n>-t2")
      assert("<dl><dt>dt</dt><dd>dd</dd></dl>", ":dt:dd")
      assert("<dl><dd>dd</dd></dl>", "::dd")
      assert("<pre>s\nt</pre>", " s\n t")
      assert("<table><tr><td>s1</td><td>s2</td></tr><tr><td>t1</td><td>t2</td></tr></table>", ",s1,s2\n,t1,t2")
      assert("<pre>s\nt\n</pre>", "{{{\ns\nt\n}}}") # u?
      assert("<plugin method=\"t\" param=\"\"></plugin>", "{{t}}")
      assert("<plugin method=\"t\" param=\"a\"></plugin>", "{{t(a)}}")
      assert("<plugin method=\"t\" param=\"\">s</plugin>", "{{t\ns\n}}")
      assert("<plugin method=\"t\" param=\"\">s1\ns2</plugin>", "{{t\ns1\ns2\n}}")
      assert("<p>t</p>", "t")
      assert("<p>s\nt</p>", "s\nt")
      assert("<p>s</p>\n<p>t</p>", "s\n\nt")

      # test_inline
      assert("<p>t</p>", "t")
      assert("<p><em>t</em></p>", "''t''")
      assert("<p><a href=\"t.html\">t</a></p>", "[[t]]")
      assert("<p><a href=\"1.html\">1</a></p>", "[[1]]")
      assert("<p><plugin method=\"interwiki\" param=\"Test:t\">Test:t</plugin></p>", "[[Test:t]]")
      assert("<plugin method=\"t\" param=\"\"></plugin>", "{{t}}")
      assert("<p>a<plugin method=\"t\" param=\"\"></plugin>b</p>", "a{{t}}b")
      assert("<p><a href=\"http://e.com/\">http://e.com/</a></p>", "http://e.com/")
      assert("<p>a <a href=\"http://e.com/\">http://e.com/</a> b</p>", "a http://e.com/ b")

      # test_multiline
      assert("<p>s\nt</p>", "s\nt")
      assert("<p>s<br/>t</p>", "s~\nt~")

      # test_qwik_formatter
      assert("", "")
      assert("<h2>t</h2>", "*t")

      # test_formatter
      # test_header
      assert("<h2>t</h2>", "*t")
      assert("<h3>t</h3>", "**t")
      assert("<h4>t</h4>", "***t")

      # test_ignore_space
      assert("<h2>t</h2>", "* t")
      assert("<h2>t</h2>", "*t ")
      assert("<h2>t</h2>", "* t ")
      assert("<h3>t</h3>", "** t")
      assert("<h3>t</h3>", "**t ")
      assert("<h3>t</h3>", "** t ")

      # test_escape
      assert("<h2>'</h2>", "*'")
      assert("<h2>&lt;</h2>", "*<")

      # test_sjis
      assert("<h2></h2>", "* ")
      assert("<h3></h3>", "** ")
      assert("<h3></h3>", "** ")
      assert("<h3></h3>", "** ")
      assert("<h3></h3>", "** ")

      # test ?
      assert("<h2>a</h2>", "*a")
      assert("<h2>a</h2>", "* a")
      assert("<h2>a</h2>", "*  a")
      assert("<h2><a href=\"t.html\">t</a></h2>", "*[[t]]")
      assert("<h2><a href=\"t.html\">t</a></h2>", "* [[t]]")
      assert("<h2><a href=\"t.html\">t</a></h2>", "* [[t]] ")
      assert("<h2><a href=\"t.html\">t</a> a</h2>", "* [[t]] a")
      assert("<h2><a href=\"t.html\">t</a>  a</h2>", "* [[t]]  a")

      # test_format
      assert("<h2>t</h2>", "*t")
      assert("<h2>&lt;</h2>", "*<")
      assert("<p><strong></strong></p>", "''''''")
    end
  end
end
